unit ColoredCheckListBoxes;

{$mode objfpc}{$H+}
{$modeswitch advancedrecords}
{$modeswitch typehelpers}

interface

uses
  // RTL, FCL
  Classes, SysUtils, Types, fgl,
  // LCL
  CheckLst, StdCtrls, Graphics, Themes;

type

  TColorList = class(specialize TFPGList<TColor>);

  { TColoredCheckListBox }

  TColoredCheckListBox = class(TCheckListBox)
  private
    FCheckBoxWidth: Integer;
    FColors: TColorList;
    FItemRect: TRect;
    function GetItemColor(AIndex: Integer): TColor;
    procedure SetItemColor(AIndex: Integer; AValue: TColor);
  protected
    procedure DrawCheckBox(AChecked: Boolean);
    procedure DrawColoredText(AIndex: Integer);
    procedure DrawItem(AIndex: Integer; ARect: TRect; AState: TOwnerDrawState); override;
    function  GetCheckBoxWidth: Integer;
    procedure SetStyle(Val: TListBoxStyle); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddItem(const Item: String; AColor: TColor; AnObject: TObject = nil); overload;
    procedure Clear; override;
  public
    property ItemColor[AIndex: Integer]: TColor read GetItemColor write SetItemColor;
  end;

implementation

const
  SIndexOutOfRange = '%s: Index[%d] out of range! Max[%d] allowed!';

{ TColoredCheckListBox }

function TColoredCheckListBox.GetItemColor(AIndex: Integer): TColor;
begin
  if AIndex > Items.Count - 1 then
    raise Exception.Create(Format(SIndexOutOfRange, ['TColoredCheckListBox.GetItemColor', AIndex, Items.Count - 1]));
  while AIndex > FColors.Count - 1 do FColors.Add(clDefault);
  Result := FColors[AIndex];
end;

procedure TColoredCheckListBox.SetItemColor(AIndex: Integer; AValue: TColor);
begin
  if AIndex > Items.Count - 1 then
    raise Exception.Create(Format(SIndexOutOfRange, ['TColoredCheckListBox.SetItemColor', AIndex, Items.Count - 1]));
  while AIndex > FColors.Count - 1 do FColors.Add(clDefault);
  FColors[AIndex] := AValue;
end;

procedure TColoredCheckListBox.DrawCheckBox(AChecked: Boolean);
var
  CheckBoxRect: TRect;
  CheckBoxDetails: TThemedElementDetails;
begin
  CheckBoxRect := FItemRect;

  if UseRightToLeftAlignment then
    CheckBoxRect.Left := CheckBoxRect.Right - FCheckBoxWidth
  else
    CheckBoxRect.Width := FCheckBoxWidth;
  Canvas.FillRect(CheckBoxRect);

  if AChecked then
    if Enabled then CheckBoxDetails := ThemeServices.GetElementDetails(tbCheckBoxCheckedNormal)
               else CheckBoxDetails := ThemeServices.GetElementDetails(tbCheckBoxCheckedDisabled)
  else
    if Enabled then CheckBoxDetails := ThemeServices.GetElementDetails(tbCheckBoxUncheckedNormal)
               else CheckBoxDetails := ThemeServices.GetElementDetails(tbCheckBoxUncheckedDisabled);
  ThemeServices.DrawElement(Canvas.Handle, CheckBoxDetails, CheckBoxRect, nil);
end;

procedure TColoredCheckListBox.DrawColoredText(AIndex: Integer);
var
  AColor, OldColor: TColor;
begin
  Canvas.FillRect(FItemRect);
  OldColor := Canvas.Font.Color;
  AColor := ItemColor[AIndex];
  if AColor <> clDefault then
    Canvas.Font.Color := AColor;
  Canvas.TextRect(FItemRect, FItemRect.Left + 2, FItemRect.Top, Items[AIndex]);
  Canvas.Font.Color := OldColor;
end;

procedure TColoredCheckListBox.DrawItem(AIndex: Integer; ARect: TRect; AState: TOwnerDrawState);
begin
  if FCheckBoxWidth < 0 then
    FCheckBoxWidth := GetCheckBoxWidth;

  FItemRect := ARect;
  DrawCheckBox(Checked[AIndex]);

  if not Header[AIndex] then
  begin
    if UseRightToLeftAlignment then
      Dec(FItemRect.Right, FCheckBoxWidth)
    else
      Inc(FItemRect.Left, FCheckBoxWidth);
  end;

  if Assigned(OnDrawItem) then
    OnDrawItem(Self, AIndex, FItemRect, AState)
  else
    DrawColoredText(AIndex);
end;

function TColoredCheckListBox.GetCheckBoxWidth: Integer;
var
  CheckBoxDetails: TThemedElementDetails;
  ASize: TSize;
begin
  Result := GetCheckWidth;
  if Result <= 0 then
  begin
    CheckBoxDetails := ThemeServices.GetElementDetails(tbCheckBoxCheckedDisabled);
    ASize := ThemeServices.GetDetailSize(CheckBoxDetails);
    Result := ASize.Width;
  end;
  if Result <= 0 then
    Result := 16;        // if nothing works, take 16 as default
  Result := Result + 4;  // add 2 px space around checkbox
end;

procedure TColoredCheckListBox.SetStyle(Val: TListBoxStyle);
begin
  inherited SetStyle(lbOwnerDrawFixed);
end;

constructor TColoredCheckListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FColors := TColorList.Create;
  FCheckBoxWidth := -1;
  SetStyle(lbOwnerDrawFixed);
end;

destructor TColoredCheckListBox.Destroy;
begin
  FColors.Free;
  inherited Destroy;
end;

procedure TColoredCheckListBox.AddItem(const Item: String; AColor: TColor; AnObject: TObject);
begin
  AddItem(Item, AnObject);
  ItemColor[Items.Count - 1] := AColor;
end;

procedure TColoredCheckListBox.Clear;
begin
  inherited Clear;
  FColors.Clear;
end;

end.

